# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: messages.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 33, 0, "", "messages.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0emessages.proto\x12\x0bsaengra_api"*\n\x06Vertex\x12\x11\n\ttype_name\x18\x01 \x01(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c"Y\n\x04\x45\x64ge\x12!\n\x04\x66rom\x18\x01 \x01(\x0b\x32\x13.saengra_api.Vertex\x12\r\n\x05label\x18\x02 \x01(\x0c\x12\x1f\n\x02to\x18\x03 \x01(\x0b\x32\x13.saengra_api.Vertex"\x86\x01\n\x08Position\x12#\n\x06vertex\x18\x01 \x01(\x0b\x32\x13.saengra_api.Vertex\x12\x30\n\x04kind\x18\x02 \x01(\x0e\x32".saengra_api.Position.PositionKind"#\n\x0cPositionKind\x12\x08\n\x04\x43ORE\x10\x00\x12\t\n\x05ORBIT\x10\x01"\xa1\x02\n\x08Subgraph\x12-\n\x0estart_position\x18\x01 \x01(\x0b\x32\x15.saengra_api.Position\x12,\n\rend_positions\x18\x02 \x03(\x0b\x32\x15.saengra_api.Position\x12%\n\x08vertices\x18\x03 \x03(\x0b\x32\x13.saengra_api.Vertex\x12 \n\x05\x65\x64ges\x18\x04 \x03(\x0b\x32\x11.saengra_api.Edge\x12-\n\x04refs\x18\x05 \x03(\x0b\x32\x1f.saengra_api.Subgraph.RefsEntry\x1a@\n\tRefsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12"\n\x05value\x18\x02 \x01(\x0b\x32\x13.saengra_api.Vertex:\x02\x38\x01"\xb4\x02\n\x07Request\x12\'\n\x07\x63onnect\x18\x01 \x01(\x0b\x32\x14.saengra_api.ConnectH\x00\x12\x32\n\rapply_updates\x18\x02 \x01(\x0b\x32\x19.saengra_api.ApplyUpdatesH\x00\x12%\n\x06\x63ommit\x18\x03 \x01(\x0b\x32\x13.saengra_api.CommitH\x00\x12)\n\x08rollback\x18\x04 \x01(\x0b\x32\x15.saengra_api.RollbackH\x00\x12\'\n\x07observe\x18\x05 \x01(\x0b\x32\x14.saengra_api.ObserveH\x00\x12!\n\x04\x66ind\x18\x06 \x01(\x0b\x32\x11.saengra_api.FindH\x00\x12#\n\x05match\x18\x07 \x01(\x0b\x32\x12.saengra_api.MatchH\x00\x42\t\n\x07request"\x1b\n\x07\x43onnect\x12\x10\n\x08graph_id\x18\x01 \x01(\t"\xde\x02\n\x0c\x41pplyUpdates\x12\x31\n\x07updates\x18\x01 \x03(\x0b\x32 .saengra_api.ApplyUpdates.Update\x1a\x9a\x02\n\x06Update\x12\x39\n\x04kind\x18\x01 \x01(\x0e\x32+.saengra_api.ApplyUpdates.Update.UpdateKind\x12!\n\x04\x66rom\x18\x02 \x01(\x0b\x32\x13.saengra_api.Vertex\x12\x12\n\x05label\x18\x03 \x01(\tH\x00\x88\x01\x01\x12$\n\x02to\x18\x04 \x01(\x0b\x32\x13.saengra_api.VertexH\x01\x88\x01\x01"g\n\nUpdateKind\x12\x0e\n\nADD_VERTEX\x10\x00\x12\x0c\n\x08\x41\x44\x44_EDGE\x10\x01\x12\x11\n\rREMOVE_VERTEX\x10\x02\x12\x0f\n\x0bREMOVE_EDGE\x10\x03\x12\x17\n\x13REMOVE_EDGES_TO_ALL\x10\x04\x42\x08\n\x06_labelB\x05\n\x03_to"\x08\n\x06\x43ommit"\n\n\x08Rollback"\xd2\x01\n\x07Observe\x12\x30\n\tobservers\x18\x01 \x03(\x0b\x32\x1d.saengra_api.Observe.Observer\x1a\x94\x01\n\x08Observer\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\nexpression\x18\x02 \x01(\t\x12/\n\x12placeholder_values\x18\x03 \x03(\x0b\x32\x13.saengra_api.Vertex\x12\x11\n\ton_create\x18\x04 \x01(\x08\x12\x11\n\ton_change\x18\x05 \x01(\x08\x12\x11\n\ton_delete\x18\x06 \x01(\x08"\xc3\x01\n\x04\x46ind\x12\x10\n\x08vertices\x18\x01 \x01(\x08\x12\x1b\n\x0ewith_type_name\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\r\n\x05\x65\x64ges\x18\x03 \x01(\x08\x12\x17\n\nwith_label\x18\x04 \x01(\tH\x01\x88\x01\x01\x12!\n\x04\x66rom\x18\x06 \x03(\x0b\x32\x13.saengra_api.Vertex\x12\x1f\n\x02to\x18\x05 \x03(\x0b\x32\x13.saengra_api.VertexB\x11\n\x0f_with_type_nameB\r\n\x0b_with_label"L\n\x05Match\x12\x12\n\nexpression\x18\x01 \x01(\t\x12/\n\x12placeholder_values\x18\x02 \x03(\x0b\x32\x13.saengra_api.Vertex"\xee\x02\n\x08Response\x12/\n\x07\x63onnect\x18\x01 \x01(\x0b\x32\x1c.saengra_api.ConnectResponseH\x00\x12:\n\rapply_updates\x18\x02 \x01(\x0b\x32!.saengra_api.ApplyUpdatesResponseH\x00\x12-\n\x06\x63ommit\x18\x03 \x01(\x0b\x32\x1b.saengra_api.CommitResponseH\x00\x12\x31\n\x08rollback\x18\x04 \x01(\x0b\x32\x1d.saengra_api.RollbackResponseH\x00\x12/\n\x07observe\x18\x05 \x01(\x0b\x32\x1c.saengra_api.ObserveResponseH\x00\x12)\n\x04\x66ind\x18\x06 \x01(\x0b\x32\x19.saengra_api.FindResponseH\x00\x12+\n\x05match\x18\x07 \x01(\x0b\x32\x1a.saengra_api.MatchResponseH\x00\x42\n\n\x08response"|\n\x0f\x43onnectResponse\x12\x33\n\x06status\x18\x01 \x01(\x0e\x32#.saengra_api.ConnectResponse.Status\x12\x0f\n\x07\x63reated\x18\x02 \x01(\x08"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"u\n\x14\x41pplyUpdatesResponse\x12\x38\n\x06status\x18\x01 \x01(\x0e\x32(.saengra_api.ApplyUpdatesResponse.Status"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"\x83\x04\n\x0e\x43ommitResponse\x12\x32\n\x06status\x18\x01 \x01(\x0e\x32".saengra_api.CommitResponse.Status\x12<\n\x0bobservation\x18\x02 \x03(\x0b\x32\'.saengra_api.CommitResponse.Observation\x1a\xbb\x02\n\x0bObservation\x12\x13\n\x0bobserver_id\x18\x01 \x01(\t\x12\x45\n\x04type\x18\x02 \x01(\x0e\x32\x37.saengra_api.CommitResponse.Observation.ObservationType\x12I\n\tvariables\x18\x03 \x03(\x0b\x32\x36.saengra_api.CommitResponse.Observation.VariablesEntry\x1a\x45\n\x0eVariablesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12"\n\x05value\x18\x02 \x01(\x0b\x32\x13.saengra_api.Vertex:\x02\x38\x01">\n\x0fObservationType\x12\r\n\tON_CREATE\x10\x00\x12\r\n\tON_CHANGE\x10\x01\x12\r\n\tON_DELETE\x10\x02"A\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01\x12\x1c\n\x18REQUIRES_REPEATED_COMMIT\x10\x02"m\n\x10RollbackResponse\x12\x34\n\x06status\x18\x01 \x01(\x0e\x32$.saengra_api.RollbackResponse.Status"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"k\n\x0fObserveResponse\x12\x33\n\x06status\x18\x01 \x01(\x0e\x32#.saengra_api.ObserveResponse.Status"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"\xae\x01\n\x0c\x46indResponse\x12\x30\n\x06status\x18\x01 \x01(\x0e\x32 .saengra_api.FindResponse.Status\x12%\n\x08vertices\x18\x02 \x03(\x0b\x32\x13.saengra_api.Vertex\x12 \n\x05\x65\x64ges\x18\x03 \x03(\x0b\x32\x11.saengra_api.Edge"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01"\x91\x01\n\rMatchResponse\x12\x31\n\x06status\x18\x01 \x01(\x0e\x32!.saengra_api.MatchResponse.Status\x12(\n\tsubgraphs\x18\x02 \x03(\x0b\x32\x15.saengra_api.Subgraph"#\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\x11\n\rNOT_CONNECTED\x10\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "messages_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_SUBGRAPH_REFSENTRY"]._loaded_options = None
    _globals["_SUBGRAPH_REFSENTRY"]._serialized_options = b"8\001"
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._loaded_options = None
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._serialized_options = (
        b"8\001"
    )
    _globals["_VERTEX"]._serialized_start = 31
    _globals["_VERTEX"]._serialized_end = 73
    _globals["_EDGE"]._serialized_start = 75
    _globals["_EDGE"]._serialized_end = 164
    _globals["_POSITION"]._serialized_start = 167
    _globals["_POSITION"]._serialized_end = 301
    _globals["_POSITION_POSITIONKIND"]._serialized_start = 266
    _globals["_POSITION_POSITIONKIND"]._serialized_end = 301
    _globals["_SUBGRAPH"]._serialized_start = 304
    _globals["_SUBGRAPH"]._serialized_end = 593
    _globals["_SUBGRAPH_REFSENTRY"]._serialized_start = 529
    _globals["_SUBGRAPH_REFSENTRY"]._serialized_end = 593
    _globals["_REQUEST"]._serialized_start = 596
    _globals["_REQUEST"]._serialized_end = 904
    _globals["_CONNECT"]._serialized_start = 906
    _globals["_CONNECT"]._serialized_end = 933
    _globals["_APPLYUPDATES"]._serialized_start = 936
    _globals["_APPLYUPDATES"]._serialized_end = 1286
    _globals["_APPLYUPDATES_UPDATE"]._serialized_start = 1004
    _globals["_APPLYUPDATES_UPDATE"]._serialized_end = 1286
    _globals["_APPLYUPDATES_UPDATE_UPDATEKIND"]._serialized_start = 1166
    _globals["_APPLYUPDATES_UPDATE_UPDATEKIND"]._serialized_end = 1269
    _globals["_COMMIT"]._serialized_start = 1288
    _globals["_COMMIT"]._serialized_end = 1296
    _globals["_ROLLBACK"]._serialized_start = 1298
    _globals["_ROLLBACK"]._serialized_end = 1308
    _globals["_OBSERVE"]._serialized_start = 1311
    _globals["_OBSERVE"]._serialized_end = 1521
    _globals["_OBSERVE_OBSERVER"]._serialized_start = 1373
    _globals["_OBSERVE_OBSERVER"]._serialized_end = 1521
    _globals["_FIND"]._serialized_start = 1524
    _globals["_FIND"]._serialized_end = 1719
    _globals["_MATCH"]._serialized_start = 1721
    _globals["_MATCH"]._serialized_end = 1797
    _globals["_RESPONSE"]._serialized_start = 1800
    _globals["_RESPONSE"]._serialized_end = 2166
    _globals["_CONNECTRESPONSE"]._serialized_start = 2168
    _globals["_CONNECTRESPONSE"]._serialized_end = 2292
    _globals["_CONNECTRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_CONNECTRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_APPLYUPDATESRESPONSE"]._serialized_start = 2294
    _globals["_APPLYUPDATESRESPONSE"]._serialized_end = 2411
    _globals["_APPLYUPDATESRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_APPLYUPDATESRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_COMMITRESPONSE"]._serialized_start = 2414
    _globals["_COMMITRESPONSE"]._serialized_end = 2929
    _globals["_COMMITRESPONSE_OBSERVATION"]._serialized_start = 2547
    _globals["_COMMITRESPONSE_OBSERVATION"]._serialized_end = 2862
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._serialized_start = 2729
    _globals["_COMMITRESPONSE_OBSERVATION_VARIABLESENTRY"]._serialized_end = 2798
    _globals["_COMMITRESPONSE_OBSERVATION_OBSERVATIONTYPE"]._serialized_start = 2800
    _globals["_COMMITRESPONSE_OBSERVATION_OBSERVATIONTYPE"]._serialized_end = 2862
    _globals["_COMMITRESPONSE_STATUS"]._serialized_start = 2864
    _globals["_COMMITRESPONSE_STATUS"]._serialized_end = 2929
    _globals["_ROLLBACKRESPONSE"]._serialized_start = 2931
    _globals["_ROLLBACKRESPONSE"]._serialized_end = 3040
    _globals["_ROLLBACKRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_ROLLBACKRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_OBSERVERESPONSE"]._serialized_start = 3042
    _globals["_OBSERVERESPONSE"]._serialized_end = 3149
    _globals["_OBSERVERESPONSE_STATUS"]._serialized_start = 2257
    _globals["_OBSERVERESPONSE_STATUS"]._serialized_end = 2292
    _globals["_FINDRESPONSE"]._serialized_start = 3152
    _globals["_FINDRESPONSE"]._serialized_end = 3326
    _globals["_FINDRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_FINDRESPONSE_STATUS"]._serialized_end = 2292
    _globals["_MATCHRESPONSE"]._serialized_start = 3329
    _globals["_MATCHRESPONSE"]._serialized_end = 3474
    _globals["_MATCHRESPONSE_STATUS"]._serialized_start = 2257
    _globals["_MATCHRESPONSE_STATUS"]._serialized_end = 2292
# @@protoc_insertion_point(module_scope)
